module kernels

    use iso_fortran_env
    use omp_lib
#ifdef MPI
    use mpi
#endif

contains

    function pi_cpu(numtrials, numthreads, numprocs, procid) result(pi0)

        implicit none
        integer(int64), intent(in):: numtrials
        integer, intent(in):: numthreads, numprocs, procid
        real(real64):: pi, r(2), pi0, pi_loc
        integer(int64):: i, nin, numtrials_loc
        integer(int32):: ierr
        
        call omp_set_num_threads(numthreads)

        numtrials_loc = ceiling(numtrials/real(numprocs))
        if (procid == numprocs-1) numtrials_loc = numtrials - procid*numtrials_loc

        pi = 0.d0

        !$OMP PARALLEL DEFAULT(PRIVATE) SHARED(pi, procid, numtrials_loc, numtrials, pi0)
        nin = 0
        pi0 = 0.d0
        !$OMP DO
        do i = 1, numtrials_loc

            call random_number(r)
            if (r(1)*r(1) + r(2)*r(2) <= 1.d0) nin = nin + 1

        end do
        !$OMP END DO NOWAIT
        pi_loc = (4.d0*nin)/numtrials

        !$OMP ATOMIC
        pi = pi + pi_loc

        !$OMP BARRIER

        !$OMP MASTER
#ifdef MPI
        call MPI_Reduce(pi, pi0, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, MPI_COMM_WORLD, ierr)
#else
        pi0 = pi
#endif
        !$OMP END MASTER
        !$OMP END PARALLEL

    end function pi_cpu

end module kernels

program main

    use kernels

    implicit none
    logical:: run
    integer:: i, nargs, numthreads, procid, numprocs, ierr, provided_thread_level, numreps
    integer(int64):: numtrials
    character(32):: arg1, arg2
    double precision:: pi, tstart
    character(*), parameter:: errmsg = "There was a problem parsing arguments!", &
        helpmsg = "Usage:"//NEW_LINE('a')//"    pi-cpu [options]"//NEW_LINE('a')//NEW_LINE('a')//"Options:"&
                //NEW_LINE('a')//"    --parallel <N>, -p <N>    Number of threads to use (default: 1)"&
                //NEW_LINE('a')//"    --trials <N>, -n <N>      Number of trials to calculate pi (default: 123,456,789)"&
                //NEW_LINE('a')//"    --reps <N>, -r <N>        Number of times to calculate pi (default: 5)"&
                //NEW_LINE('a')//"where <N> is an integer."

#ifdef MPI
    call MPI_Init_thread(MPI_THREAD_FUNNELED, provided_thread_level, ierr)

    call MPI_Comm_size(MPI_COMM_WORLD, numprocs, ierr)
    call MPI_Comm_rank(MPI_COMM_WORLD, procid, ierr)
#else
    procid = 0
    numprocs = 1
#endif

    ! setting defaults
    numthreads = 1
    numtrials = 123456789
    numreps = 5
    ierr = 0

    nargs = command_argument_count()

    run = .true.

    do i = 1, nargs, 2
        call get_command_argument(i, arg1)
        call get_command_argument(i+1, arg2)
        select case (arg1)
        case ("-p", "--parallel")
            read(arg2, *, iostat=ierr) numthreads
        case ("-n", "--trials")
            read(arg2, *, iostat=ierr) numtrials
        case ("-r", "--repetitions")
            read(arg2, *, iostat=ierr) numreps
        case ("-h", "--help")
            write(output_unit, "(A)") helpmsg
            stop
        case default
            ierr = 1
        end select
    end do

    if (ierr /= 0) then
        write(error_unit, "(A)") errmsg
        write(error_unit, "(A)") helpmsg
        stop
    end if

    !do i = 1, numreps
    i = 0
    do while (i /= numreps)
        i = i + 1
        tstart = omp_get_wtime()
        pi = pi_cpu(numtrials, numthreads, numprocs, procid)
        if (procid == 0) then
            write(output_unit, '(A, f15.13, 1x, A, f16.13, 1x, A, f6.4, A)') &
                "Result: ", pi, &
                "Error: ", pi - 3.14159265358979311599796346854, &
                "Time: ", omp_get_wtime() - tstart, "s"
        end if
    end do

#ifdef MPI
    call MPI_Finalize(ierr)
#endif

end program main
