#include <thrust/device_vector.h>
#include <iostream>
#include <iomanip>
#include <thrust/reduce.h>
#include <thrust/random.h>
#include <string>
#include <cmath>
#include <curand.h>
#include <chrono>

using namespace std::chrono;
using namespace std::chrono_literals;

const std::string errmsg = "There was a problem parsing arguments!\n";
const std::string helpmsg = \
"Usage:\n\
    pi-gpu [options]\n\n\
\
Options:\n\
    --parallel <N>, -p <N>    Number of GPUs to use (default: 1)\n\
    --trials <N>, -n <N>      Number of trials to calculate pi (default: 123,456,789)\n\
    --reps <N>, -r <N>        Number of times to calculate pi (default: 5)\n\
    --mode <N>, -m <N>        Mode index (default: 1, fast/experimental: 2)\n\
where <N> is an integer.\n";

// rng operator for "slow" version
struct prg
{
    double a, b;
    unsigned int N;

    __host__ __device__
    prg(double _a=0., double _b=1., unsigned int _N=1234) : a(_a), b(_b), N(_N) {};

    __host__ __device__
        double operator()(const unsigned int n) const
        {
            thrust::default_random_engine rng(N);
            thrust::uniform_real_distribution<double> dist(a, b);
            rng.discard(n);

            return dist(rng);
        }
};

// global function to calculate magnitude of vector positions
__global__
void dist(int n, const double* xy, double* r)
{
    int i = blockIdx.x*blockDim.x + threadIdx.x;
    if (i < n) r[i] = xy[2*i]*xy[2*i] + xy[2*i+1]*xy[2*i+1] < 1.; 
};

// argparser
int myargtoi(const int i, char* argv[], int& iarg)
{
    try
    {
        std::size_t pos;
        std::string sarg = argv[i];
        iarg = std::stoi(sarg, &pos);
        return 0;
    } catch (...)
    {
        return 22;
    }
}

double pi_slow(const int numgpus, const int numtrials, const int itrial)
{

    thrust::device_vector<double>* ddxy[numgpus];
    thrust::device_vector<double>* ddr[numgpus];
    int nin_total = 0;

    int numtrials_loc = ceil(static_cast<double>(numtrials)/numgpus);

    for (int dev = 0; dev < numgpus; ++dev)
    {
        if (dev == numgpus-1) numtrials_loc = numtrials - (numgpus-1)*numtrials_loc;

        cudaSetDevice(dev);
        ddxy[dev] = new thrust::device_vector<double>(2*numtrials_loc);
        ddr[dev] = new thrust::device_vector<double>(numtrials_loc);

        thrust::counting_iterator<unsigned int> index_sequence_begin(0);
        const unsigned int seed = 1234+itrial;
        thrust::transform(index_sequence_begin,
                index_sequence_begin + 2*numtrials_loc,
                (*ddxy[dev]).begin(),
                prg(0., 1., seed));

        double* xy_ptr = thrust::raw_pointer_cast((*ddxy[dev]).data());
        double* r_ptr = thrust::raw_pointer_cast((*ddr[dev]).data());
        dist<<<numtrials/1024+1, 1024>>>(numtrials_loc, xy_ptr, r_ptr);

        nin_total += thrust::reduce((*ddr[dev]).begin(), (*ddr[dev]).begin()+numtrials_loc);
        xy_ptr = NULL;
        r_ptr = NULL;
        delete ddxy[dev];
        delete ddr[dev];
    }

    return 4.*static_cast<double>(nin_total)/static_cast<double>(numtrials);
}

double pi_fast(const int numgpus, const int numtrials, const int itrial)
{

    double* ddxy[numgpus];
    double* ddr[numgpus];
    int numtrials_loc[numgpus];

    for (int dev = 0; dev < numgpus; ++dev)
    {
        numtrials_loc[dev] = ceil(static_cast<double>(numtrials)/numgpus);
        if (dev == numgpus-1) numtrials_loc[dev] = numtrials - (numgpus-1)*numtrials_loc[dev];

        cudaSetDevice(dev);

        curandGenerator_t gen;
        curandCreateGenerator(&gen, CURAND_RNG_PSEUDO_DEFAULT);

        const long long int seed = 1234 + itrial;
        curandSetPseudoRandomGeneratorSeed(gen, seed);

        cudaMalloc(&ddxy[dev], 2*numtrials_loc[dev]*sizeof(double));
        curandGenerateUniformDouble(gen, ddxy[dev], 2*numtrials_loc[dev]);

        cudaMalloc(&ddr[dev], numtrials_loc[dev]*sizeof(double));

        dist<<<numtrials/1024+1, 1024>>>(numtrials_loc[dev], ddxy[dev], ddr[dev]);

    }

    int nin_total = 0;

    for (int dev = 0; dev < numgpus; ++dev)
    {

        cudaSetDevice(dev);

        thrust::device_ptr<double> r_ptr(ddr[dev]);
        // nin_total += thrust::count_if(r_ptr, r_ptr+numtrials_loc[dev], is_in());
        nin_total += thrust::reduce(r_ptr, r_ptr+numtrials_loc[dev]);

        r_ptr = NULL;
        
        cudaFree(ddxy[dev]);
        cudaFree(ddr[dev]);
    }

    return 4.*static_cast<double>(nin_total)/static_cast<double>(numtrials);
}

double get_wtime()
{
    return duration<double>(high_resolution_clock::now().time_since_epoch()).count();
}

int main(int argc, char* argv[])
{

    int numreps = 5;
    int numtrials = 123456789;
    int numgpus=1;
    int ierr = 0;
    int mode = 1;

    for (int i = 1; i < argc; i += 2)
    {
        std::string flag = argv[i];
        if (flag == "-p" || flag == "--parallel")
        {
            ierr = myargtoi(i+1, argv, numgpus);
            int nDevices;
            cudaGetDeviceCount(&nDevices);
            if (numgpus > nDevices) ierr = 22;
        } else if (flag == "-n" || flag == "--trials")
        {
            ierr = myargtoi(i+1, argv, numtrials);
        } else if (flag == "-r" || flag == "--reps")
        {
            ierr = myargtoi(i+1, argv, numreps);
        } else if (flag == "-m" || flag == "--mode") 
        {
            ierr = myargtoi(i+1, argv, mode);
            if (mode < 1 || mode > 2) ierr = 22;
        } else if (flag == "-h" || flag == "--help")
        {
            std::cout << helpmsg;
            exit(0);
        } else
        {
            ierr = 22;
        }
    }

    if (ierr > 0) 
    {
        std::cerr << errmsg;
        exit(ierr);
    }

    // Checking numgpus
    int nDevices;
    cudaGetDeviceCount(&nDevices);

    // setting up
    double pi_est;
    double wtime = 0.;
    
    // beginning calculation of pi
    for (int i = 0; i != numreps; ++i)
    {
        wtime = -get_wtime();
        switch (mode)
        {
            case 2:
                pi_est = pi_fast(numgpus, numtrials, i);
                break;
            default:
                pi_est = pi_slow(numgpus, numtrials, i);
                break;
        }
        wtime += get_wtime();
        std::cout << "Result: " << std::fixed << std::setw(15) << std::setprecision(13)  << pi_est << \
        " Error: " << std::setw(16) << pi_est - 3.14159265358979311599796346854 << \
        std::setw(6) << std::setprecision(4) << " Time: " << wtime << "s\n";
    }

    return 0;
}
